

local Loc = LibStub ("AceLocale-3.0"):GetLocale ("Details_DeathGraphs")
local SharedMedia = LibStub:GetLibrary("LibSharedMedia-3.0")

--> create the plugin object
	local DeathGraphs = _detalhes:NewPluginObject ("Details_DeathGraphs", DETAILSPLUGIN_ALWAYSENABLED)
	tinsert (UISpecialFrames, "Details_DeathGraphs")
--> main frame (shortcut)
	local DeathGraphsFrame = DeathGraphs.Frame
 
DeathGraphs.version_string = "v1.5.0"

local CONST_DBTYPE_DEATH = "deaths"
local CONST_DBTYPE_ENDURANCE = "endurance"

DeathGraphs:SetPluginDescription ("Stores all deaths during raid boss encounters and calculates the player value to the group based on how often the player dies.")

local debugmode = false

local function CreatePluginFunctions()

	function DeathGraphs:DebugMsg (msg, sec)
		if (debugmode) then
			DeathGraphs:Msg (msg, sec)
		end
	end

	function DeathGraphs:OpenWindow()
		if (not DeathGraphs.frames_built) then
			_detalhes.DeathGraphsWindowBuilder (DeathGraphs)
			_detalhes.DeathGraphsWindowBuilder = nil
			DeathGraphs.frames_built = true
			
			local current_segment = DeathGraphs:GetCurrentCombat()
			if (current_segment and current_segment.is_boss and current_segment.is_boss.diff and current_segment.is_boss.id) then
				DeathGraphs.db.last_boss = DeathGraphs.last_encounter_hash or ("" .. current_segment.is_boss.id .. current_segment.is_boss.diff)
			end
		end
		
		DeathGraphs:Refresh()
		DeathGraphsFrame:Show()
	end
	
	function DeathGraphs:CloseWindow()
		DeathGraphsFrame:Hide()
	end
	
	DeathGraphs.ToolbarButton = DeathGraphs.ToolBar:NewPluginToolbarButton (DeathGraphs.OpenWindow, "Interface\\AddOns\\Details_DeathGraphs\\icon", "A.D.L", Loc ["STRING_TOOLTIP"], 16, 16, "DEATHGRAPHICS_BUTTON")
	DeathGraphs.ToolbarButton.shadow = true
	
	function DeathGraphs:CanShowIcon()
		if (self.db.show_icon == 1) then
			local found_something = false
			
			for _, boss in pairs (DeathGraphs.deaths_database) do
				if (boss) then
					found_something = true
					break
				end
			end
			
			if (not found_something) then
				for _, boss in pairs (DeathGraphs.endurance_database) do
					if (boss) then
						found_something = true
						break
					end
				end
			end
			
			if (found_something) then
				DeathGraphs:ShowToolbarIcon (DeathGraphs.ToolbarButton, "star")
			else
				DeathGraphs:HideToolbarIcon (DeathGraphs.ToolbarButton)
			end
		end
	end
	
	function DeathGraphs:HideIcon()
		DeathGraphs:HideToolbarIcon (DeathGraphs.ToolbarButton)
	end

	function DeathGraphs:OnDetailsEvent (event, ...)
		if (event == "HIDE") then --> plugin hidded, disabled
			self.open = false
		
		elseif (event == "SHOW") then --> plugin hidded, disabled
			self.open = true
			
			if (not DeathGraphs.db.first_run) then
			
				DeathGraphs.db.first_run = true
			
				--> first run
				local welcome = CreateFrame ("frame", nil, DeathGraphsFrame)
				welcome:SetFrameStrata ("TOOLTIP")
				welcome:SetPoint ("center", DeathGraphsFrame, "center")
				welcome:SetSize (400, 250)
				DeathGraphs.welcome_panel = welcome
				welcome:Show()
				welcome:SetBackdrop ({edgeFile = "Interface\\Buttons\\UI-SliderBar-Border", edgeSize = 8,
				bgFile = [[Interface\AddOns\Details\images\background]], tile = true, tileSize = 130, insets = {left = 1, right = 1, top = 5, bottom = 5}})
				
				local str = _detalhes.gump:CreateLabel (welcome, "Welcome to Advanced Death Logs!\n\n\n- Select the mode on the bottom left dropdown box:\n\n - |cFFFFFF00Deaths|r: show common death logs plus the overall damage taken before the player's death.\n\n - |cFFFFFF00Endurance|r: get the first deaths (set under 'Config Death Limits') and assign 90 points for those player, all other players get 100 points. After few tries in a encounter, you may see which players are dying more frequently.\n\n\n- You can always close the window clicking with the right mouse button.", 
				nil, nil, "GameFontNormal")
				str:SetPoint (15, -15)
				str:SetWidth (370)
				
				local close_button = _detalhes.gump:CreateButton (welcome, function() welcome:Hide() end, 86, 16, "Close")
				close_button:InstallCustomTexture()
				close_button:SetPoint ("center", welcome, "center")
				close_button:SetPoint ("bottom", welcome, "bottom", 0, 10)
			end
			
		elseif (event == "COMBAT_PLAYER_ENTER") then --> combat started
			--DeathGraphs:ScheduleTimer ("ResetSurvivalTable", 1)
			--_combat_object = select (1, ...)
		
		elseif (event == "COMBAT_PLAYER_LEAVE") then --> combat ended
			DeathGraphs:DebugMsg ("combat finished -> calling CombatFinished()")
			DeathGraphs:CombatFinished (...)
			--DeathGraphs:CloseSurvivalTable()
			--DeathGraphs:WasBossEncounter()
			
		elseif (event == "DETAILS_DATA_RESET") then
			--DeathGraphs:ResetData()
			
		elseif (event == "PLUGIN_DISABLED") then
			DeathGraphs:HideIcon()
			
		elseif (event == "PLUGIN_ENABLED" or event == "DETAILS_STARTED") then
			DeathGraphs:CanShowIcon()

		end
	end

	function DeathGraphs:GetPlayerTable (boss_table, player_name)
	
		local player_table = boss_table.player_db [player_name]
		if (player_table) then
			return player_table
		end
		
		if (boss_table.type == CONST_DBTYPE_DEATH) then
			local t = {
				overall = {},
				deaths = {},
				name = player_name,
				class = select (2, UnitClass (player_name))
			}
			boss_table.player_db [player_name] = t
			return t
			
		elseif (boss_table.type == CONST_DBTYPE_ENDURANCE) then
			local t = {
				points = 0,
				encounters = 0,
				deaths = {},
				class = select (2, UnitClass (player_name))
			}
			boss_table.player_db [player_name] = t
			return t
			
		end
	end

	--[[	boss table structure:
		DeathGraphs.deaths_database = {
			[ENCOUNTER_ID + DIFFICULT ID] = {
				player_db = {
					[player name] = {data}
				},
				type = "deaths" or "endurance"
			}
		} --]]
	
	function DeathGraphs:GetBossTable (encounter_id, boss, type)
	
		local db_table
		if (type == CONST_DBTYPE_DEATH) then
			db_table = DeathGraphs.deaths_database
		elseif (type == CONST_DBTYPE_ENDURANCE) then
			db_table = DeathGraphs.endurance_database
		end
	
		local hash
		if (boss) then
			hash = "" .. encounter_id .. boss.diff
		else
			hash = "" .. encounter_id
		end
	
		local boss_table = db_table [hash]
		if (boss_table) then
			return boss_table
		end
	
		local t = {
			player_db = {},
			id = encounter_id,
			name = boss.name,
			diff = boss.diff,
			hash = hash,
			boss_table = table_deepcopy (boss)
		}
		
		if (type == CONST_DBTYPE_DEATH) then
			t.type = CONST_DBTYPE_DEATH
			DeathGraphs.deaths_database [hash] = t
			
		elseif (type == CONST_DBTYPE_ENDURANCE) then
			t.type = CONST_DBTYPE_ENDURANCE
			DeathGraphs.endurance_database [hash] = t
		end
	
		return t
	end
	
	function DeathGraphs:CombatFinished (combat)
		
		local difficult = combat:GetDifficulty()
		
		if (difficult) then
			--> normal mode
			if (difficult == 14) then
				if (not DeathGraphs.db.captures [2]) then
					DeathGraphs:DebugMsg ("Normal mode isn't active, not recording this segment.")
					return
				end
			--> heroic mode
			elseif (difficult == 15) then
				if (not DeathGraphs.db.captures [3]) then
					DeathGraphs:DebugMsg ("Heroic mode isn't active, not recording this segment.")
					return
				end
			--> raid finder
			elseif (difficult == 17) then
				if (not DeathGraphs.db.captures [1]) then
					DeathGraphs:DebugMsg ("Raid Finder mode isn't active, not recording this segment.")
					return
				end
			--> mythic
			elseif (difficult == 16) then
				if (not DeathGraphs.db.captures [4]) then
					DeathGraphs:DebugMsg ("Mythic mode isn't active, not recording this segment.")
					return
				end
			else
				return
			end
		end
		
		--> read all deaths
	
		local boss = combat:GetBossInfo()
		DeathGraphs:DebugMsg ("bossinfo", boss)
		
		DeathGraphs.db.last_combat_id = DeathGraphs.combat_id
		
		if (boss) then
		
			DeathGraphs:DebugMsg ("boss found", boss.name)
		
			local EI = DeathGraphs:GetEncounterIdFromBossIndex (boss.mapid, boss.index)
			if (EI) then
			
				DeathGraphs:DebugMsg ("boss EI: " .. EI .. " diff: " .. combat:GetDifficulty())
			
				local diff = combat:GetDifficulty()
				if (not diff) then
					return
				end
			
				DeathGraphs:DebugMsg ("" .. EI .. boss.diff)
			
				local boss_table = DeathGraphs:GetBossTable (EI, boss, CONST_DBTYPE_DEATH)
				local endurance_table = DeathGraphs:GetBossTable (EI, boss, CONST_DBTYPE_ENDURANCE)
				
				DeathGraphs.last_encounter_hash = "" .. EI .. boss.diff
				
				--> iterate beetween deaths occured in latest encounter
				local death_list = combat:GetDeaths()
				local endurance_failed = {}
				
				DeathGraphs:DebugMsg ("deaths: " .. #death_list)
				
				if (#death_list > 0) then
					
					--> get raid size
					--local size = select (5, GetInstanceInfo())
					if (combat:GetCombatTime() > 40) then
						local group_size = GetNumGroupMembers()
						local size = group_size
						local max
					
						if (size <= 14) then
							max = 3
						elseif (size >= 15 and size <= 19) then
							max = 4
						elseif (size >= 20 and size <= 24) then
							max = 5
						elseif (size >= 25) then
							max = 6
						end
						
						max = DeathGraphs.db.endurance_threshold
					
						local deaths_stored = 0
					
						for i, t in ipairs (death_list) do
						
							local _, class = UnitClass (t[3])
							local player_table = DeathGraphs:GetPlayerTable (boss_table, t[3])
							
							if (not player_table) then
								if (debugmode) then
									print ("NO player_table", t[3], boss_table.type, class)
								end
							else
							
								if (deaths_stored >= DeathGraphs.db.deaths_threshold) then
									break
								end
							
								deaths_stored = deaths_stored + 1
							
								--> store death
								local d = table_deepcopy (t[1])
								
								local last = d [#d]
								if (last [4]) then
									if (type (last[1]) == "number" and last[1] == 3) then
										tremove (d, #d)
									end
								end
								
								while (#d > 16) do
									tremove (d, 1)
								end
								
								tinsert (player_table.deaths, 1, {
									d, --death table (last events)
									t.last_cooldown or false, --last coioldown
									t.dead_at, --time of death in seconds
									t[5], --unit max health
									boss.name, 
									t[2], --time() of death
									class, --player class
									(DeathGraphs.combat_counter or 1) --combat counter
								})
								
								--> store overall
								local d_stored = 0
								local time_of_death = t[2]
								
								for i = #t[1], 1, -1 do 
								
									local damage_event = t[1][i]

									if (type (damage_event [1]) == "boolean" and damage_event [1]) then
										local time = damage_event [4]
									
										if (time+8 > time_of_death) then
											local spellid = damage_event [2]
											if (player_table.overall [spellid]) then
												player_table.overall [spellid] = player_table.overall [spellid] + damage_event [3]
											else
												player_table.overall [spellid] = damage_event [3]
											end
											
											d_stored = d_stored + 1
											if (d_stored >= 4) then
												break
											end
										end
									end
								end

							end

							--> store endurance
							if (i <= max) then
								local player_table = DeathGraphs:GetPlayerTable (endurance_table, t[3])
								
								if (endurance_failed [t[3]]) then
									player_table.points = player_table.points + 80
								else
									player_table.points = player_table.points + 90
								end
								
								player_table.encounters = player_table.encounters + 1
								
								local last_hit = DeathGraphs:GetLastHit (t[1])
								tinsert (player_table.deaths, {DeathGraphs.combat_id, t.dead_at, last_hit})
								endurance_failed [t[3]] = true
							end
						end--loop
					end --combat time > 40
				end --#death_list > 0
				
				--> close the rest of endurance
				if (combat:GetCombatTime() > 40) then
					for i = 1, GetNumGroupMembers(), 1 do
						local name = GetUnitName ("raid"..i, true)
						if (not endurance_failed [name]) then
							local player_table = DeathGraphs:GetPlayerTable (endurance_table, name)
							player_table.points = player_table.points + 100
							player_table.encounters = player_table.encounters + 1
						end
					end
				end
				
				DeathGraphs:CanShowIcon()
				
			end
		else
			DeathGraphs:DebugMsg ("boss not found on latest segment.")
		end --is boss
		
	end --combat finished

end

function DeathGraphs:GetLastHit (deathlog)
	for i = #deathlog, 1, -1 do
		local hit = deathlog[i]
		--[1] boolean (true)
		if (type (hit [1]) == "boolean" and hit [1]) then
			local spellname = DeathGraphs.getspellinfo (hit [2]) or ""
			return spellname .. " |cFFFF3333" .. DeathGraphs:comma_value (hit [3]) .. "|r"
		end
	end
	return ""
end

local build_options_panel = function()
	local options_frame = DeathGraphs:CreatePluginOptionsFrame ("DeathGraphsOptionsWindow", "Advanced Death Logs Options", 1)
	local menu = {
		{
			type = "range",
			get = function() return DeathGraphs.db.deaths_threshold end,
			set = function (self, fixedparam, value) DeathGraphs.db.deaths_threshold = value end,
			min = 1,
			max = 30,
			step = 1,
			desc = "Adjust the amount of deaths to capture on each encounter try.\n\nThis is important due to synergy with |cFFFFFF00Overall Damage Taken Before Death|r and Wipe Calls by the Raid Leader.",
			name = "Deaths Threshold",
		},
		{
			type = "range",
			get = function() return DeathGraphs.db.endurance_threshold end,
			set = function (self, fixedparam, value) DeathGraphs.db.endurance_threshold = value end,
			min = 1,
			max = 30,
			step = 1,
			desc = "The first |cFFFFFF00X|r players to die receive 90 endurance points, all others receive 100.",
			name = "Endurance Threshold",
		},
	}
	
	DeathGraphs:GetFramework():BuildMenu (options_frame, menu, 15, -75, 260)
	options_frame:SetBackdropColor (0, 0, 0, .9)
end

DeathGraphs.OpenOptionsPanel = function()
	if (not DeathGraphsOptionsWindow) then
		build_options_panel()
	end
	DeathGraphsOptionsWindow:Show()
end

function DeathGraphs:OnEvent (_, event, ...)

	if (event == "ADDON_LOADED") then
		local AddonName = select (1, ...)
		if (AddonName == "Details_DeathGraphs") then
			
			if (_G._detalhes) then
				
				--> catch data
				DeathGraphs.deaths_database = DeathGraphsDBDeaths or {}
				DeathGraphsDBDeaths = DeathGraphs.deaths_database
				
				DeathGraphs.endurance_database = DeathGraphsDBEndurance or {}
				DeathGraphsDBEndurance = DeathGraphs.endurance_database
				
				--> check old versions without endurance amount
				for bossid, bosstable in pairs (DeathGraphs.endurance_database) do
					for playername, playertable in pairs (bosstable.player_db) do
						if (not playertable.encounters) then
							DeathGraphs.endurance_database [bossid].player_db [playername] = nil
						end
					end
				end
				
			--	DeathGraphs.endurance_database -> hashboss, bosstable -> player_db -> playername, playertable -> encounters
			--	/run DeathGraphsDBEndurance=nil;DeathGraphsDBDeaths=nil
				
				--> create widgets
				CreatePluginFunctions()

				--> core version required
				local MINIMAL_DETAILS_VERSION_REQUIRED = 28
				
				local defaults = {
					show_icon = 1,
					last_boss = false,
					last_player = false,
					last_segment = false,
					last_encounter_hash = false,
					showing_type = 1,
					last_combat_id = 0,
					captures = {false, false, true, true},
					first_run = false,
					deaths_threshold = 10,
					endurance_threshold = 5,
				}
				
				--> Install
				local install, saveddata = _G._detalhes:InstallPlugin ("TOOLBAR", Loc ["STRING_PLUGIN_NAME"], "Interface\\AddOns\\Details_DeathGraphs\\icon", DeathGraphs, "DETAILS_PLUGIN_DEATH_GRAPHICS", MINIMAL_DETAILS_VERSION_REQUIRED, "Details! Team", DeathGraphs.version_string, defaults)
				if (type (install) == "table" and install.error) then
					print (install.error)
				end
				
				--> Register needed events
				_G._detalhes:RegisterEvent (DeathGraphs, "DETAILS_DATA_RESET")
				_G._detalhes:RegisterEvent (DeathGraphs, "COMBAT_PLAYER_LEAVE")

			end
		end
	end
end
